# This Python file uses the following encoding: utf-8
#
# Copyright 2016 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Tests for the places module."""

import uuid

from types import GeneratorType

import responses

import neshan
import neshan.test as _test


class PlacesTest(_test.TestCase):

    def setUp(self):
        self.key = 'service.8AwKYt4Tov5II9uvbK5BkiZZ1TMHAM4JnBtm4yft'
        self.client = neshan.Client(self.key)
        self.location = (36.268706, 59.610011)

    @responses.activate
    def test_places_text_search(self):
        url = 'https://api.neshan.org/v1/search'
        responses.add(responses.GET, url,
                      body='{"items": []}',
                      status=200, content_type='application/json;charset=UTF-8')

        self.client.places('رستوران', location=self.location)

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('%s?lat=36.268706&lng=59.610011&term=رستوران'
                            % (url), responses.calls[0].request.url)



#
# Copyright 2015 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Performs requests to the Neshan API."""

import neshan
from neshan import convert


def snap_to_roads(client, path):
    """Snaps a path to the most likely roads travelled.

    Takes points collected along a route, and returns a similar
    set of data with the points snapped to the most likely roads the vehicle
    was traveling along.

    :param path: The path to be snapped.
    :type path: a single location, or a list of locations, where a
        location is a string, dict, list, or tuple

    :rtype: A list of snapped points.
    """

    params = {"path": convert.location_list(path)}

    return client._request("/v1/map-matching", params).get("snappedPoints", [])



#
# Copyright 2014 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Performs requests to the Neshan Geocoding API."""
from neshan import convert


# def geocode(client, address=None, components=None, bounds=None, region=None,
#             language=None):
#     """
#     Geocoding is the process of converting addresses
#     (like ``"1600 Amphitheatre Parkway, Mountain View, CA"``) into geographic
#     coordinates (like latitude 37.423021 and longitude -122.083739), which you
#     can use to place markers or position the map.

#     :param address: The address to geocode.
#     :type address: string

#     :param components: A component filter for which you wish to obtain a
#         geocode, for example: ``{'administrative_area': 'TX','country': 'US'}``
#     :type components: dict

#     :param bounds: The bounding box of the viewport within which to bias geocode
#         results more prominently.
#     :type bounds: string or dict with northeast and southwest keys.

#     :param region: The region code, specified as a ccTLD ("top-level domain")
#         two-character value.
#     :type region: string

#     :param language: The language in which to return results.
#     :type language: string

#     :rtype: list of geocoding results.
#     """

#     params = {}

#     if address:
#         params["address"] = address

#     if components:
#         params["components"] = convert.components(components)

#     if bounds:
#         params["bounds"] = convert.bounds(bounds)

#     if region:
#         params["region"] = region

#     if language:
#         params["language"] = language

#     return client._request("/maps/api/geocode/json", params).get("results", [])


def reverse_geocode(client, latlng):
    """
    Reverse geocoding is the process of converting geographic coordinates into a
    human-readable address.

    :param latlng: The latitude/longitude value for which you wish
        to obtain the closest, human-readable address.
    :type latlng: dict, list, or tuple

    :rtype: list of reverse geocoding results.
    """
    latlng = convert.normalize_lat_lng(latlng)
    params = {
        "lat": latlng[0],
        "lng": latlng[1],
    }

    return client._request("/v2/reverse", params)



#
# Copyright 2014 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Tests for the directions module."""

from datetime import datetime
from datetime import timedelta
import time

import responses

import neshan
import neshan.test as _test


class DirectionsTest(_test.TestCase):

    def setUp(self):
        self.key = 'service.8AwKYt4Tov5II9uvbK5BkiZZ1TMHAM4JnBtm4yft'
        self.client = neshan.Client(self.key)

    @responses.activate
    def test_simple_directions(self):
        responses.add(responses.GET,
                      'https://api.neshan.org/v2/direction',
                      body='{"routes":[]}',
                      status=200,
                      content_type='application/json;charset=UTF-8')

        # Simplest directions request. Driving directions by default.
        routes = self.client.directions(
            (36.268706, 59.610011), (36.287607, 59.599527))

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.neshan.org/v2/direction'
                            '?origin=36.268706,59.610011&destination=36.287607,59.599527',
                            responses.calls[0].request.url)

    # @responses.activate
    # def test_zero_results_returns_response(self):
    #     responses.add(responses.GET,
    #                   'https://api.neshan.org/v2/direction',
    #                   body='{"status":"ZERO_RESULTS","routes":[]}',
    #                   status=200,
    #                   content_type='application/json;charset=UTF-8')

    #     routes = self.client.directions("Toledo", "Madrid")
    #     self.assertIsNotNone(routes)
    #     self.assertEqual(0, len(routes))

    @responses.activate
    def test_alternatives(self):
        responses.add(responses.GET,
                      'https://api.neshan.org/v2/direction',
                      body='{"routes":[]}',
                      status=200,
                      content_type='application/json;charset=UTF-8')

        routes = self.client.directions((36.268706, 59.610011),
                                        (36.287607, 59.599527),
                                        alternatives=True)

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.neshan.org/v2/direction?'
                            'origin=36.268706,59.610011&destination=36.287607,59.599527&'
                            'alternative=true',
                            responses.calls[0].request.url)



#
# Copyright 2014 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

import unittest
import codecs

try: # Python 3
    from urllib.parse import urlparse, parse_qsl
except ImportError: # Python 2
    from urlparse import urlparse, parse_qsl


class TestCase(unittest.TestCase):

    def assertURLEqual(self, first, second, msg=None):
        """Check that two arguments are equivalent URLs. Ignores the order of
        query arguments.
        """
        first_parsed = urlparse(first)
        second_parsed = urlparse(second)
        self.assertEqual(first_parsed[:3], second_parsed[:3], msg)

        first_qsl = sorted(parse_qsl(first_parsed.query))
        second_qsl = sorted(parse_qsl(second_parsed.query))
        self.assertEqual(first_qsl, second_qsl, msg)

    def u(self, string):
        """Create a unicode string, compatible across all versions of Python."""
        # NOTE(cbro): Python 3-3.2 does not have the u'' syntax.
        return codecs.unicode_escape_decode(string)[0]



#
# Copyright 2014 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#


"""Tests for client module."""

import time

import responses
import requests
import uuid

import neshan
import neshan.client as _client
import neshan.test as _test
from neshan.client import _X_GOOG_MAPS_EXPERIENCE_ID


class ClientTest(_test.TestCase):

    def test_invalid_api_key(self):
        with self.assertRaises(Exception):
            client = neshan.Client(key="Invalid key.")
            client.directions("Sydney", "Melbourne")

    def test_urlencode(self):
        # See GH #72.
        encoded_params = _client.urlencode_params([("address", "=Sydney ~")])
        self.assertEqual("address=%3DSydney+~", encoded_params)

    # @responses.activate
    # def test_queries_per_second(self):
    #     # This test assumes that the time to run a mocked query is
    #     # relatively small, eg a few milliseconds. We define a rate of
    #     # 3 queries per second, and run double that, which should take at
    #     # least 1 second but no more than 2.
    #     queries_per_second = 3
    #     query_range = range(queries_per_second * 2)
    #     for _ in query_range:
    #         responses.add(responses.GET,
    #                       "https://maps.googleapis.com/maps/api/geocode/json",
    #                       body='{"status":"OK","results":[]}',
    #                       status=200,
    #                       content_type="application/json")
    #     client = neshan.Client(key="AIzaasdf",
    #                                queries_per_second=queries_per_second)
    #     start = time.time()
    #     for _ in query_range:
    #         client.geocode("Sesame St.")
    #     end = time.time()
    #     self.assertTrue(start + 1 < end < start + 2)

    # @responses.activate
    # def test_key_sent(self):
    #     responses.add(responses.GET,
    #                   "https://maps.googleapis.com/maps/api/geocode/json",
    #                   body='{"status":"OK","results":[]}',
    #                   status=200,
    #                   content_type="application/json")

    #     client = neshan.Client(key="AIzaasdf")
    #     client.geocode("Sesame St.")

    #     self.assertEqual(1, len(responses.calls))
    #     self.assertURLEqual("https://maps.googleapis.com/maps/api/geocode/json?"
    #                         "key=AIzaasdf&address=Sesame+St.",
    #                         responses.calls[0].request.url)

    # @responses.activate
    # def test_extra_params(self):
    #     responses.add(responses.GET,
    #                   "https://maps.googleapis.com/maps/api/geocode/json",
    #                   body='{"status":"OK","results":[]}',
    #                   status=200,
    #                   content_type="application/json")

    #     client = neshan.Client(key="AIzaasdf")
    #     client.geocode("Sesame St.", extra_params={"foo": "bar"})

    #     self.assertEqual(1, len(responses.calls))
    #     self.assertURLEqual("https://maps.googleapis.com/maps/api/geocode/json?"
    #                         "key=AIzaasdf&address=Sesame+St.&foo=bar",
    #                         responses.calls[0].request.url)

    # def test_hmac(self):
    #     """
    #     From http://en.wikipedia.org/wiki/Hash-based_message_authentication_code

    #     HMAC_SHA1("key", "The quick brown fox jumps over the lazy dog")
    #        = 0xde7c9b85b8b78aa6bc8a7a36f70a90701c9db4d9
    #     """

    #     message = "The quick brown fox jumps over the lazy dog"
    #     key = "a2V5" # "key" -> base64
    #     signature = "3nybhbi3iqa8ino29wqQcBydtNk="

    #     self.assertEqual(signature, _client.sign_hmac(key, message))

    # @responses.activate
    # def test_url_signed(self):
    #     responses.add(responses.GET,
    #                   "https://maps.googleapis.com/maps/api/geocode/json",
    #                   body='{"status":"OK","results":[]}',
    #                   status=200,
    #                   content_type="application/json")

    #     client = neshan.Client(client_id="foo", client_secret="a2V5")
    #     client.geocode("Sesame St.")

    #     self.assertEqual(1, len(responses.calls))

    #     # Check ordering of parameters.
    #     self.assertIn("address=Sesame+St.&client=foo&signature",
    #             responses.calls[0].request.url)
    #     self.assertURLEqual("https://maps.googleapis.com/maps/api/geocode/json?"
    #                         "address=Sesame+St.&client=foo&"
    #                         "signature=fxbWUIcNPZSekVOhp2ul9LW5TpY=",
    #                         responses.calls[0].request.url)

    # @responses.activate
    # def test_ua_sent(self):
    #     responses.add(responses.GET,
    #                   "https://maps.googleapis.com/maps/api/geocode/json",
    #                   body='{"status":"OK","results":[]}',
    #                   status=200,
    #                   content_type="application/json")

    #     client = neshan.Client(key="AIzaasdf")
    #     client.geocode("Sesame St.")

    #     self.assertEqual(1, len(responses.calls))
    #     user_agent = responses.calls[0].request.headers["User-Agent"]
    #     self.assertTrue(user_agent.startswith("GoogleGeoApiClientPython"))

    # @responses.activate
    # def test_retry(self):
    #     class request_callback:
    #         def __init__(self):
    #             self.first_req = True

    #         def __call__(self, req):
    #             if self.first_req:
    #                 self.first_req = False
    #                 return (200, {}, '{"status":"OVER_QUERY_LIMIT"}')
    #             return (200, {}, '{"status":"OK","results":[]}')

    #     responses.add_callback(responses.GET,
    #             "https://maps.googleapis.com/maps/api/geocode/json",
    #             content_type='application/json',
    #             callback=request_callback())

    #     client = neshan.Client(key="AIzaasdf")
    #     client.geocode("Sesame St.")

    #     self.assertEqual(2, len(responses.calls))
    #     self.assertEqual(responses.calls[0].request.url, responses.calls[1].request.url)

    # @responses.activate
    # def test_transport_error(self):
    #     responses.add(responses.GET,
    #                   "https://maps.googleapis.com/maps/api/geocode/json",
    #                   status=404,
    #                   content_type='application/json')

    #     client = neshan.Client(key="AIzaasdf")
    #     with self.assertRaises(neshan.exceptions.HTTPError) as e:
    #         client.geocode("Foo")

    #     self.assertEqual(e.exception.status_code, 404)

    # @responses.activate
    # def test_retry_intermittent(self):
    #     class request_callback:
    #         def __init__(self):
    #             self.first_req = True

    #         def __call__(self, req):
    #             if self.first_req:
    #                 self.first_req = False
    #                 return (500, {}, 'Internal Server Error.')
    #             return (200, {}, '{"status":"OK","results":[]}')

    #     responses.add_callback(responses.GET,
    #             "https://maps.googleapis.com/maps/api/geocode/json",
    #             content_type="application/json",
    #             callback=request_callback())

    #     client = neshan.Client(key="AIzaasdf")
    #     client.geocode("Sesame St.")

    #     self.assertEqual(2, len(responses.calls))

    # def test_channel_without_client_id(self):
    #     with self.assertRaises(ValueError):
    #         client = neshan.Client(key="AIzaasdf", channel="mychannel")

    def test_requests_version(self):
        client_args_timeout = {
            "key": "service.8AwKYt4Tov5II9uvbK5BkiZZ1TMHAM4JnBtm4yft",
            "connect_timeout": 5,
            "read_timeout": 5
        }
        client_args = client_args_timeout.copy()
        del client_args["connect_timeout"]
        del client_args["read_timeout"]

        requests.__version__ = '2.3.0'
        with self.assertRaises(NotImplementedError):
            neshan.Client(**client_args_timeout)
        neshan.Client(**client_args)

        requests.__version__ = '2.4.0'
        neshan.Client(**client_args_timeout)
        neshan.Client(**client_args)

    # @responses.activate
    # def _perform_mock_request(self, experience_id=None):
    #     # Mock response
    #     responses.add(responses.GET,
    #                   "https://maps.googleapis.com/maps/api/geocode/json",
    #                   body='{"status":"OK","results":[]}',
    #                   status=200,
    #                   content_type="application/json")

    #     # Perform network call
    #     client = neshan.Client(key="AIzaasdf")
    #     client.set_experience_id(experience_id)
    #     client.geocode("Sesame St.")
    #     return responses.calls[0].request

    # @responses.activate
    # def test_no_retry_over_query_limit(self):
    #     responses.add(responses.GET,
    #                   "https://maps.googleapis.com/foo",
    #                   body='{"status":"OVER_QUERY_LIMIT"}',
    #                   status=200,
    #                   content_type="application/json")

    #     client = neshan.Client(key="AIzaasdf",
    #                                retry_over_query_limit=False)

    #     with self.assertRaises(neshan.exceptions.ApiError):
    #         client._request("/foo", {})

    #     self.assertEqual(1, len(responses.calls))



#
# Copyright 2015 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Tests for the roads module."""


import responses

import neshan
import neshan.test as _test


class RoadsTest(_test.TestCase):

    def setUp(self):
        self.key = "service.8AwKYt4Tov5II9uvbK5BkiZZ1TMHAM4JnBtm4yft"
        self.client = neshan.Client(self.key)

    @responses.activate
    def test_snap(self):
        responses.add(responses.GET,
                      "https://api.neshan.org/v1/map-matching",
                      body='{"snappedPoints":["foo"]}',
                      status=200,
                      content_type="application/json")

        results = self.client.snap_to_roads((40.714728, -73.998672))
        self.assertEqual("foo", results[0])

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual("https://api.neshan.org/v1/map-matching?"
                            "path=40.714728%2C-73.998672",
                            responses.calls[0].request.url)

    # @responses.activate
    # def test_retry(self):
    #     class request_callback:
    #         def __init__(self):
    #             self.first_req = True

    #         def __call__(self, req):
    #             if self.first_req:
    #                 self.first_req = False
    #                 return (500, {}, 'Internal Server Error.')
    #             return (200, {}, '{"speedLimits":[]}')

    #     responses.add_callback(responses.GET,
    #             "https://roads.googleapis.com/v1/speedLimits",
    #             content_type="application/json",
    #             callback=request_callback())

    #     self.client.speed_limits([])

    #     self.assertEqual(2, len(responses.calls))
    #     self.assertEqual(responses.calls[0].request.url, responses.calls[1].request.url)



#
# Copyright 2014 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Tests for the distance matrix module."""

from datetime import datetime
import time

import responses

import neshan
import neshan.test as _test


class DistanceMatrixTest(_test.TestCase):

    def setUp(self):
        self.key = 'service.8AwKYt4Tov5II9uvbK5BkiZZ1TMHAM4JnBtm4yft'
        self.client = neshan.Client(self.key)

    @responses.activate
    def test_basic_params(self):
        responses.add(responses.GET,
                      'https://api.neshan.org/v1/distance-matrix',
                      body='{"status":"OK","rows":[]}',
                      status=200,
                      content_type='application/json')

        origins = ['36.316551,59.585658', '35.691080,51.299734']
        destinations = ['32.633881,51.681525','31.876915,54.344318']

        matrix = self.client.distance_matrix(origins, destinations)

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.neshan.org/v1/distance-matrix?'
                            'origins=36.316551,59.585658|35.691080,51.299734&'
                            'destinations=32.633881,51.681525|31.876915,54.344318',
                            responses.calls[0].request.url)



#
# Copyright 2015 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Performs requests to the Neshan API."""
import warnings

from neshan import convert


def places(
    client,
    term,
    location=None,
):
    """
    Places search.

    :param term: The text string on which to search, for example: "restaurant".
    :type term: string

    :param location: The latitude/longitude value for which you wish to obtain the
        closest, human-readable address.
    :type location: dict, list, or tuple

    :rtype: result dict with the following keys:
        count: number of returned results
        items: array of places
    """
    params = {}
    if location:
        latlng = convert.normalize_lat_lng(location)
        params['lat'] = latlng[0]
        params['lng'] = latlng[1]
    if term:
        params['term'] = term

    url = '/v1/search'
    return client._request(url, params)



# This Python file uses the following encoding: utf-8
#
# Copyright 2014 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Tests for the geocoding module."""

import datetime

import responses

import neshan
import neshan.test as _test


class GeocodingTest(_test.TestCase):

    def setUp(self):
        self.key = 'service.8AwKYt4Tov5II9uvbK5BkiZZ1TMHAM4JnBtm4yft'
        self.client = neshan.Client(self.key)

    # @responses.activate
    # def test_simple_geocode(self):
    #     responses.add(responses.GET,
    #                   'https://api.neshan.org/v2/reverse',
    #                   body='{"status":"OK","results":[]}',
    #                   status=200,
    #                   content_type='application/json')

    #     results = self.client.geocode('Sydney')

    #     self.assertEqual(1, len(responses.calls))
    #     self.assertURLEqual('https://api.neshan.org/v2/reverse?'
    #                         'key=%s&address=Sydney' % self.key,
    #                         responses.calls[0].request.url)

    @responses.activate
    def test_reverse_geocode(self):
        responses.add(responses.GET,
                      'https://api.neshan.org/v2/reverse',
                      body='{"status":"OK","results":[]}',
                      status=200,
                      content_type='application/json')

        results = self.client.reverse_geocode((-33.8674869, 151.2069902))

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.neshan.org/v2/reverse?'
                            'lat=-33.8674869&lng=151.2069902',
                            responses.calls[0].request.url)

    @responses.activate
    def test_simple_reverse_geocode(self):
        responses.add(responses.GET,
                      'https://api.neshan.org/v2/reverse',
                      body='{"status":"OK","results":[]}',
                      status=200,
                      content_type='application/json')

        results = self.client.reverse_geocode((40.714224, -73.961452))

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.neshan.org/v2/reverse?'
                            'lat=40.714224&lng=-73.961452',
                            responses.calls[0].request.url)



#
# Copyright 2014 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Converts Python types to string representations suitable for Maps API server.

    For example:

    sydney = {
        "lat" : -33.8674869,
        "lng" : 151.2069902
    }

    convert.latlng(sydney)
    # '-33.8674869,151.2069902'
"""

import time as _time


def format_float(arg):
    """Formats a float value to be as short as possible.

    Truncates float to 8 decimal places and trims extraneous
    trailing zeros and period to give API args the best
    possible chance of fitting within 2000 char URL length
    restrictions.

    For example:

    format_float(40) -> "40"
    format_float(40.0) -> "40"
    format_float(40.1) -> "40.1"
    format_float(40.001) -> "40.001"
    format_float(40.0010) -> "40.001"
    format_float(40.000000001) -> "40"
    format_float(40.000000009) -> "40.00000001"

    :param arg: The lat or lng float.
    :type arg: float

    :rtype: string
    """
    return ("%.8f" % float(arg)).rstrip("0").rstrip(".")


def latlng(arg):
    """Converts a lat/lon pair to a comma-separated string.

    For example:

    sydney = {
        "lat" : -33.8674869,
        "lng" : 151.2069902
    }

    convert.latlng(sydney)
    # '-33.8674869,151.2069902'

    For convenience, also accepts lat/lon pair as a string, in
    which case it's returned unchanged.

    :param arg: The lat/lon pair.
    :type arg: string or dict or list or tuple
    """
    if is_string(arg):
        return arg

    normalized = normalize_lat_lng(arg)
    return "%s,%s" % (format_float(normalized[0]), format_float(normalized[1]))


def normalize_lat_lng(arg):
    """Take the various lat/lng representations and return a tuple.

    Accepts various representations:
    1) dict with two entries - "lat" and "lng"
    2) list or tuple - e.g. (-33, 151) or [-33, 151]

    :param arg: The lat/lng pair.
    :type arg: dict or list or tuple

    :rtype: tuple (lat, lng)
    """
    if isinstance(arg, dict):
        if "lat" in arg and "lng" in arg:
            return arg["lat"], arg["lng"]
        if "latitude" in arg and "longitude" in arg:
            return arg["latitude"], arg["longitude"]

    # List or tuple.
    if _is_list(arg):
        return arg[0], arg[1]

    raise TypeError(
        "Expected a lat/lng dict or tuple, "
        "but got %s" % type(arg).__name__)


def location_list(arg):
    """Joins a list of locations into a pipe separated string, handling
    the various formats supported for lat/lng values.

    For example:
    p = [{"lat" : -33.867486, "lng" : 151.206990}, "Sydney"]
    convert.waypoint(p)
    # '-33.867486,151.206990|Sydney'

    :param arg: The lat/lng list.
    :type arg: list

    :rtype: string
    """
    if isinstance(arg, tuple):
        # Handle the single-tuple lat/lng case.
        return latlng(arg)
    else:
        return "|".join([latlng(location) for location in as_list(arg)])


def join_list(sep, arg):
    """If arg is list-like, then joins it with sep.

    :param sep: Separator string.
    :type sep: string

    :param arg: Value to coerce into a list.
    :type arg: string or list of strings

    :rtype: string
    """
    return sep.join(as_list(arg))


def as_list(arg):
    """Coerces arg into a list. If arg is already list-like, returns arg.
    Otherwise, returns a one-element list containing arg.

    :rtype: list
    """
    if _is_list(arg):
        return arg
    return [arg]


def _is_list(arg):
    """Checks if arg is list-like. This excludes strings and dicts."""
    if isinstance(arg, dict):
        return False
    if isinstance(arg, str): # Python 3-only, as str has __iter__
        return False
    return (not _has_method(arg, "strip")
            and _has_method(arg, "__getitem__")
            or _has_method(arg, "__iter__"))


def is_string(val):
    """Determines whether the passed value is a string, safe for 2/3."""
    try:
        basestring
    except NameError:
        return isinstance(val, str)
    return isinstance(val, basestring)


def time(arg):
    """Converts the value into a unix time (seconds since unix epoch).

    For example:
        convert.time(datetime.now())
        # '1409810596'

    :param arg: The time.
    :type arg: datetime.datetime or int
    """
    # handle datetime instances.
    if _has_method(arg, "timetuple"):
        arg = _time.mktime(arg.timetuple())

    if isinstance(arg, float):
        arg = int(arg)

    return str(arg)


def _has_method(arg, method):
    """Returns true if the given object has a method with the given name.

    :param arg: the object

    :param method: the method name
    :type method: string

    :rtype: bool
    """
    return hasattr(arg, method) and callable(getattr(arg, method))


def components(arg):
    """Converts a dict of components to the format expected by the Google Maps
    server.

    For example:
    c = {"country": "US", "postal_code": "94043"}
    convert.components(c)
    # 'country:US|postal_code:94043'

    :param arg: The component filter.
    :type arg: dict

    :rtype: basestring
    """

    # Components may have multiple values per type, here we
    # expand them into individual key/value items, eg:
    # {"country": ["US", "AU"], "foo": 1} -> "country:AU", "country:US", "foo:1"
    def expand(arg):
        for k, v in arg.items():
            for item in as_list(v):
                yield "%s:%s" % (k, item)

    if isinstance(arg, dict):
        return "|".join(sorted(expand(arg)))

    raise TypeError(
        "Expected a dict for components, "
        "but got %s" % type(arg).__name__)


def bounds(arg):
    """Converts a lat/lon bounds to a comma- and pipe-separated string.

    Accepts two representations:
    1) string: pipe-separated pair of comma-separated lat/lon pairs.
    2) dict with two entries - "southwest" and "northeast". See convert.latlng
    for information on how these can be represented.

    For example:

    sydney_bounds = {
        "northeast" : {
            "lat" : -33.4245981,
            "lng" : 151.3426361
        },
        "southwest" : {
            "lat" : -34.1692489,
            "lng" : 150.502229
        }
    }

    convert.bounds(sydney_bounds)
    # '-34.169249,150.502229|-33.424598,151.342636'

    :param arg: The bounds.
    :type arg: dict
    """

    if is_string(arg) and arg.count("|") == 1 and arg.count(",") == 2:
        return arg
    elif isinstance(arg, dict):
        if "southwest" in arg and "northeast" in arg:
            return "%s|%s" % (latlng(arg["southwest"]),
                              latlng(arg["northeast"]))

    raise TypeError(
        "Expected a bounds (southwest/northeast) dict, "
        "but got %s" % type(arg).__name__)


def decode_polyline(polyline):
    """Decodes a Polyline string into a list of lat/lng dicts.

    See the developer docs for a detailed description of this encoding:
    https://developers.google.com/maps/documentation/utilities/polylinealgorithm

    :param polyline: An encoded polyline
    :type polyline: string

    :rtype: list of dicts with lat/lng keys
    """
    points = []
    index = lat = lng = 0

    while index < len(polyline):
        result = 1
        shift = 0
        while True:
            b = ord(polyline[index]) - 63 - 1
            index += 1
            result += b << shift
            shift += 5
            if b < 0x1f:
                break
        lat += (~result >> 1) if (result & 1) != 0 else (result >> 1)

        result = 1
        shift = 0
        while True:
            b = ord(polyline[index]) - 63 - 1
            index += 1
            result += b << shift
            shift += 5
            if b < 0x1f:
                break
        lng += ~(result >> 1) if (result & 1) != 0 else (result >> 1)

        points.append({"lat": lat * 1e-5, "lng": lng * 1e-5})

    return points


def encode_polyline(points):
    """Encodes a list of points into a polyline string.

    See the developer docs for a detailed description of this encoding:
    https://developers.google.com/maps/documentation/utilities/polylinealgorithm

    :param points: a list of lat/lng pairs
    :type points: list of dicts or tuples

    :rtype: string
    """
    last_lat = last_lng = 0
    result = ""

    for point in points:
        ll = normalize_lat_lng(point)
        lat = int(round(ll[0] * 1e5))
        lng = int(round(ll[1] * 1e5))
        d_lat = lat - last_lat
        d_lng = lng - last_lng

        for v in [d_lat, d_lng]:
            v = ~(v << 1) if v < 0 else v << 1
            while v >= 0x20:
                result += (chr((0x20 | (v & 0x1f)) + 63))
                v >>= 5
            result += (chr(v + 63))

        last_lat = lat
        last_lng = lng

    return result


def shortest_path(locations):
    """Returns the shortest representation of the given locations.

    The Elevations API limits requests to 2000 characters, and accepts
    multiple locations either as pipe-delimited lat/lng values, or
    an encoded polyline, so we determine which is shortest and use it.

    :param locations: The lat/lng list.
    :type locations: list

    :rtype: string
    """
    if isinstance(locations, tuple):
        # Handle the single-tuple lat/lng case.
        locations = [locations]
    encoded = "enc:%s" % encode_polyline(locations)
    unencoded = location_list(locations)
    if len(encoded) < len(unencoded):
        return encoded
    else:
        return unencoded



#
# Copyright 2014 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Tests for the convert module."""

import datetime
import unittest
import pytest

from neshan import convert


class ConvertTest(unittest.TestCase):

    def test_latlng(self):
        expected = "1,2"
        ll = {"lat": 1, "lng": 2}
        self.assertEqual(expected, convert.latlng(ll))

        ll = [1, 2]
        self.assertEqual(expected, convert.latlng(ll))

        ll = (1, 2)
        self.assertEqual(expected, convert.latlng(ll))

        self.assertEqual(expected, convert.latlng(expected))

        with self.assertRaises(TypeError):
            convert.latlng(1)

    def test_location_list(self):
        expected = "1,2|1,2"
        ll = [{"lat": 1, "lng": 2}, {"lat": 1, "lng": 2}]
        self.assertEqual(expected, convert.location_list(ll))

        ll = [[1, 2], [1, 2]]
        self.assertEqual(expected, convert.location_list(ll))

        ll = [(1, 2), [1, 2]]
        self.assertEqual(expected, convert.location_list(ll))

        self.assertEqual(expected, convert.location_list(expected))

        with self.assertRaises(TypeError):
            convert.latlng(1)

    def test_join_list(self):
        self.assertEqual("asdf", convert.join_list("|", "asdf"))

        self.assertEqual("1,2,A", convert.join_list(",", ["1", "2", "A"]))

        self.assertEqual("", convert.join_list(",", []))

        self.assertEqual("a,B", convert.join_list(",", ("a", "B")))

    def test_as_list(self):
        self.assertEqual([1], convert.as_list(1))

        self.assertEqual([1, 2, 3], convert.as_list([1, 2, 3]))

        self.assertEqual(["string"], convert.as_list("string"))

        self.assertEqual((1, 2), convert.as_list((1, 2)))

        a_dict = {"a": 1}
        self.assertEqual([a_dict], convert.as_list(a_dict))

    def test_time(self):
        self.assertEqual("1409810596", convert.time(1409810596))

        dt = datetime.datetime.fromtimestamp(1409810596)
        self.assertEqual("1409810596", convert.time(dt))

    def test_components(self):
        c = {"country": "US"}
        self.assertEqual("country:US", convert.components(c))

        c = {"country": "US", "foo": 1}
        self.assertEqual("country:US|foo:1", convert.components(c))

        c = {"country": ["US", "AU"], "foo": 1}
        self.assertEqual("country:AU|country:US|foo:1", convert.components(c))

        with self.assertRaises(TypeError):
            convert.components("test")

        with self.assertRaises(TypeError):
            convert.components(1)

        with self.assertRaises(TypeError):
            convert.components(("c", "b"))

    def test_bounds(self):
        ne = {"lat": 1, "lng": 2}
        sw = (3, 4)
        b = {"northeast": ne, "southwest": sw}
        self.assertEqual("3,4|1,2", convert.bounds(b))

        with self.assertRaises(TypeError):
            convert.bounds("test")

    def test_polyline_decode(self):
        syd_mel_route = ("rvumEis{y[`NsfA~tAbF`bEj^h{@{KlfA~eA~`AbmEghAt~D|e@j"
                         "lRpO~yH_\\v}LjbBh~FdvCxu@`nCplDbcBf_B|wBhIfhCnqEb~D~"
                         "jCn_EngApdEtoBbfClf@t_CzcCpoEr_Gz_DxmAphDjjBxqCviEf}"
                         "B|pEvsEzbE~qGfpExjBlqCx}BvmLb`FbrQdpEvkAbjDllD|uDldD"
                         "j`Ef|AzcEx_Gtm@vuI~xArwD`dArlFnhEzmHjtC~eDluAfkC|eAd"
                         "hGpJh}N_mArrDlr@h|HzjDbsAvy@~~EdTxpJje@jlEltBboDjJdv"
                         "KyZpzExrAxpHfg@pmJg[tgJuqBnlIarAh}DbN`hCeOf_IbxA~uFt"
                         "|A|xEt_ArmBcN|sB|h@b_DjOzbJ{RlxCcfAp~AahAbqG~Gr}AerA"
                         "`dCwlCbaFo]twKt{@bsG|}A~fDlvBvz@tw@rpD_r@rqB{PvbHek@"
                         "vsHlh@ptNtm@fkD[~xFeEbyKnjDdyDbbBtuA|~Br|Gx_AfxCt}Cj"
                         "nHv`Ew\\lnBdrBfqBraD|{BldBxpG|]jqC`mArcBv]rdAxgBzdEb"
                         "{InaBzyC}AzaEaIvrCzcAzsCtfD~qGoPfeEh]h`BxiB`e@`kBxfA"
                         "v^pyA`}BhkCdoCtrC~bCxhCbgEplKrk@tiAteBwAxbCwuAnnCc]b"
                         "{FjrDdjGhhGzfCrlDruBzSrnGhvDhcFzw@n{@zxAf}Fd{IzaDnbD"
                         "joAjqJjfDlbIlzAraBxrB}K~`GpuD~`BjmDhkBp{@r_AxCrnAjrC"
                         "x`AzrBj{B|r@~qBbdAjtDnvCtNzpHxeApyC|GlfM`fHtMvqLjuEt"
                         "lDvoFbnCt|@xmAvqBkGreFm~@hlHw|AltC}NtkGvhBfaJ|~@riAx"
                         "uC~gErwCttCzjAdmGuF`iFv`AxsJftD|nDr_QtbMz_DheAf~Buy@"
                         "rlC`i@d_CljC`gBr|H|nAf_Fh{G|mE~kAhgKviEpaQnu@zwAlrA`"
                         "G~gFnvItz@j{Cng@j{D{]`tEftCdcIsPz{DddE~}PlnE|dJnzG`e"
                         "G`mF|aJdqDvoAwWjzHv`H`wOtjGzeXhhBlxErfCf{BtsCjpEjtD|"
                         "}Aja@xnAbdDt|ErMrdFh{CzgAnlCnr@`wEM~mE`bA`uD|MlwKxmB"
                         "vuFlhB|sN`_@fvBp`CxhCt_@loDsS|eDlmChgFlqCbjCxk@vbGxm"
                         "CjbMba@rpBaoClcCk_DhgEzYdzBl\\vsA_JfGztAbShkGtEhlDzh"
                         "C~w@hnB{e@yF}`D`_Ayx@~vGqn@l}CafC")

        points = convert.decode_polyline(syd_mel_route)
        self.assertAlmostEqual(-33.86746, points[0]["lat"])
        self.assertAlmostEqual(151.207090, points[0]["lng"])
        self.assertAlmostEqual(-37.814130, points[-1]["lat"])
        self.assertAlmostEqual(144.963180, points[-1]["lng"])

    def test_polyline_round_trip(self):
        test_polyline = ("gcneIpgxzRcDnBoBlEHzKjBbHlG`@`IkDxIi"
                         "KhKoMaLwTwHeIqHuAyGXeB~Ew@fFjAtIzExF")

        points = convert.decode_polyline(test_polyline)
        actual_polyline = convert.encode_polyline(points)
        self.assertEqual(test_polyline, actual_polyline)


@pytest.mark.parametrize(
    "value, expected",
    [
        (40, "40"),
        (40.0, "40"),
        (40.1, "40.1"),
        (40.00000001, "40.00000001"),
        (40.000000009, "40.00000001"),
        (40.000000001, "40"),
    ],
)
def test_format_float(value, expected):
    assert convert.format_float(value) == expected



import sys
import io
from setuptools import setup


requirements = ["requests>=2.20.0,<3.0"]

# use io.open until python2.7 support is dropped
with io.open("README.md", encoding="utf8") as f:
    readme = f.read()

with io.open("CHANGELOG.md", encoding="utf8") as f:
    changelog = f.read()


setup(
    name="neshan",
    version="1.0.3",
    author='Nima Shayanfar',
    description="Python client library for Neshan maps",
    long_description=readme + changelog,
    long_description_content_type="text/markdown",
    scripts=[],
    url="https://github.com/nshayanfar/neshan-services-python",
    packages=["neshan"],
    license="Apache 2.0",
    platforms="Posix; MacOS X; Windows",
    setup_requires=requirements,
    install_requires=requirements,
    test_suite="neshan.test",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet",
    ],
    python_requires='>=3.5',
)



#
# Copyright 2014 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Performs requests to the Google Maps Directions API."""

from neshan import convert


def directions(client, origin, destination,
               waypoints=None, avoidTrafficZone=False, avoidOddEvenZone=False,
               alternatives=False):
    """Get directions between an origin point and a destination point.

    :param origin: The address or latitude/longitude value from which you wish
        to calculate directions.
    :type origin: string, dict, list, or tuple

    :param destination: The address or latitude/longitude value from which
        you wish to calculate directions.
    :type destination: string, dict, list, or tuple

    :param waypoints: Specifies an array of waypoints. Waypoints alter a
        route by routing it through the specified location(s).
    :type waypoints: a single location, or a list of locations, where a
        location is a string, dict, list, or tuple

    :param alternatives: If True, more than one route may be returned in the
        response.
    :type alternatives: bool

    :rtype: list of routes
    """

    params = {
        "origin": convert.latlng(origin),
        "destination": convert.latlng(destination)
    }

    if waypoints:
        waypoints = convert.location_list(waypoints)
        params["waypoints"] = waypoints

    if avoidTrafficZone:
        params['avoidTrafficZone'] = 'true'

    if avoidOddEvenZone:
        params['avoidOddEvenZone'] = 'true'

    if alternatives:
        params["alternative"] = "true"

    return client._request("/v2/direction", params).get("routes", [])



#
# Copyright 2014 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""
Core client functionality, common across all API requests (including performing
HTTP requests).
"""

from neshan.places import places
from neshan.roads import snap_to_roads
from neshan.geocoding import reverse_geocode
from neshan.distance_matrix import distance_matrix
from neshan.directions import directions
import base64
import collections
from datetime import datetime
from datetime import timedelta
import functools
import hashlib
import hmac
import re
import requests
import random
import time

import neshan

try:  # Python 3
    from urllib.parse import urlencode
except ImportError:  # Python 2
    from urllib import urlencode

_X_GOOG_MAPS_EXPERIENCE_ID = "X-Goog-Maps-Experience-ID"
_USER_AGENT = "NeshanGeoApiClientPython/%s" % neshan.__version__
_DEFAULT_BASE_URL = "https://api.neshan.org"

_RETRIABLE_STATUSES = set([500, 503, 504])


class Client(object):
    """Performs requests to the Neshan API web services."""

    def __init__(self, key,
                 timeout=None, connect_timeout=None, read_timeout=None,
                 retry_timeout=60, requests_kwargs=None,
                 queries_per_second=50, channel=None,
                 retry_over_query_limit=True):
        """
        :param key: Maps API key. This is required.
        :type key: string

        :param timeout: Combined connect and read timeout for HTTP requests, in
            seconds. Specify "None" for no timeout.
        :type timeout: int

        :param connect_timeout: Connection timeout for HTTP requests, in
            seconds. You should specify read_timeout in addition to this option.
            Note that this requires requests >= 2.4.0.
        :type connect_timeout: int

        :param read_timeout: Read timeout for HTTP requests, in
            seconds. You should specify connect_timeout in addition to this
            option. Note that this requires requests >= 2.4.0.
        :type read_timeout: int

        :param retry_timeout: Timeout across multiple retriable requests, in
            seconds.
        :type retry_timeout: int

        :param queries_per_second: Number of queries per second permitted.
            If the rate limit is reached, the client will sleep for the
            appropriate amount of time before it runs the current query.
        :type queries_per_second: int

        :param retry_over_query_limit: If True, requests that result in a
            response indicating the query rate limit was exceeded will be
            retried. Defaults to True.
        :type retry_over_query_limit: bool

        :raises ValueError: when either credentials are missing, incomplete
            or invalid.
        :raises NotImplementedError: if connect_timeout and read_timeout are
            used with a version of requests prior to 2.4.0.

        :param requests_kwargs: Extra keyword arguments for the requests
            library, which among other things allow for proxy auth to be
            implemented. See the official requests docs for more info:
            http://docs.python-requests.org/en/latest/api/#main-interface
        :type requests_kwargs: dict

        """

        if key and not key.startswith("service"):
            raise ValueError("Invalid API key provided.")

        self.session = requests.Session()
        self.key = key

        if timeout and (connect_timeout or read_timeout):
            raise ValueError("Specify either timeout, or connect_timeout "
                             "and read_timeout")

        if connect_timeout and read_timeout:
            # Check that the version of requests is >= 2.4.0
            chunks = requests.__version__.split(".")
            if int(chunks[0]) < 2 or (int(chunks[0]) == 2 and int(chunks[1]) < 4):
                raise NotImplementedError("Connect/Read timeouts require "
                                          "requests v2.4.0 or higher")
            self.timeout = (connect_timeout, read_timeout)
        else:
            self.timeout = timeout

        self.retry_timeout = timedelta(seconds=retry_timeout)
        self.requests_kwargs = requests_kwargs or {}
        headers = self.requests_kwargs.pop('headers', {})
        headers.update({"User-Agent": _USER_AGENT})
        headers.update({'Api-Key': key})
        self.requests_kwargs.update({
            "headers": headers,
            "timeout": self.timeout,
            "verify": True,  # NOTE(cbro): verify SSL certs.
        })

        self.queries_per_second = queries_per_second
        self.retry_over_query_limit = retry_over_query_limit
        self.sent_times = collections.deque("", queries_per_second)

    def _request(self, url, params, first_request_time=None, retry_counter=0,
                 base_url=_DEFAULT_BASE_URL,
                 extract_body=None, requests_kwargs=None, post_json=None):
        """Performs HTTP GET/POST with credentials, returning the body as
        JSON.

        :param url: URL path for the request. Should begin with a slash.
        :type url: string

        :param params: HTTP GET parameters.
        :type params: dict or list of key/value tuples

        :param first_request_time: The time of the first request (None if no
            retries have occurred).
        :type first_request_time: datetime.datetime

        :param retry_counter: The number of this retry, or zero for first attempt.
        :type retry_counter: int

        :param base_url: The base URL for the request. Defaults to the Maps API
            server. Should not have a trailing slash.
        :type base_url: string

        :param extract_body: A function that extracts the body from the request.
            If the request was not successful, the function should raise a
            neshan.HTTPError or neshan.ApiError as appropriate.
        :type extract_body: function

        :param requests_kwargs: Same extra keywords arg for requests as per
            __init__, but provided here to allow overriding internally on a
            per-request basis.
        :type requests_kwargs: dict

        :raises ApiError: when the API returns an error.
        :raises Timeout: if the request timed out.
        :raises TransportError: when something went wrong while trying to
            exceute a request.
        """

        if not first_request_time:
            first_request_time = datetime.now()

        elapsed = datetime.now() - first_request_time
        if elapsed > self.retry_timeout:
            raise neshan.exceptions.Timeout()

        if retry_counter > 0:
            # 0.5 * (1.5 ^ i) is an increased sleep time of 1.5x per iteration,
            # starting at 0.5s when retry_counter=0. The first retry will occur
            # at 1, so subtract that first.
            delay_seconds = 0.5 * 1.5 ** (retry_counter - 1)

            # Jitter this value by 50% and pause.
            time.sleep(delay_seconds * (random.random() + 0.5))

        # Default to the client-level self.requests_kwargs, with method-level
        # requests_kwargs arg overriding.
        requests_kwargs = requests_kwargs or {}
        final_requests_kwargs = dict(self.requests_kwargs, **requests_kwargs)

        # Determine GET/POST.
        requests_method = self.session.get
        if post_json is not None:
            requests_method = self.session.post
            final_requests_kwargs["json"] = post_json

        try:
            generated_url = self._generate_url(base_url+url, params)
            response = requests_method(generated_url, **final_requests_kwargs)
        except requests.exceptions.Timeout:
            raise neshan.exceptions.Timeout()
        except Exception as e:
            raise neshan.exceptions.TransportError(e)

        if response.status_code in _RETRIABLE_STATUSES:
            # Retry request.
            return self._request(url, params, first_request_time,
                                 retry_counter + 1, base_url,
                                 extract_body, requests_kwargs, post_json)

        # Check if the time of the nth previous query (where n is
        # queries_per_second) is under a second ago - if so, sleep for
        # the difference.
        if self.sent_times and len(self.sent_times) == self.queries_per_second:
            elapsed_since_earliest = time.time() - self.sent_times[0]
            if elapsed_since_earliest < 1:
                time.sleep(1 - elapsed_since_earliest)

        try:
            if extract_body:
                result = extract_body(response)
            else:
                result = self._get_body(response)
            self.sent_times.append(time.time())
            return result
        except neshan.exceptions._RetriableRequest as e:
            if isinstance(e, neshan.exceptions._OverQueryLimit) and not self.retry_over_query_limit:
                raise

            # Retry request.
            return self._request(url, params, first_request_time,
                                 retry_counter + 1, base_url,
                                 extract_body, requests_kwargs, post_json)

    def _get(self, *args, **kwargs):  # Backwards compatibility.
        return self._request(*args, **kwargs)

    def _get_body(self, response):
        if response.status_code != 200:
            raise neshan.exceptions.HTTPError(response.status_code)

        body = response.json()

        api_status = body.get("status", None)
        if api_status and api_status.lower() != 'ok':
            code = body.get('code')
            message = body.get('message')
            raise neshan.exceptions.ApiError(code, message)

        return body


    def _generate_url(self,path, params):
        extra_params = getattr(self, "_extra_params", None) or {}
        if type(params) is dict:
            params = sorted(dict(extra_params, **params).items())
        else:
            params = sorted(extra_params.items()) + params[:]    
        path = "?".join([path, urlencode_params(params)])
        return path


def make_api_method(func):
    """
    Provides a single entry point for modifying all API methods.
    For now this is limited to allowing the client object to be modified
    with an `extra_params` keyword arg to each method, that is then used
    as the params for each web service request.

    Please note that this is an unsupported feature for advanced use only.
    It's also currently incompatibile with multiple threads, see GH #160.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        args[0]._extra_params = kwargs.pop("extra_params", None)
        result = func(*args, **kwargs)
        try:
            del args[0]._extra_params
        except AttributeError:
            pass
        return result
    return wrapper


Client.directions = make_api_method(directions)
Client.distance_matrix = make_api_method(distance_matrix)
#Client.elevation = make_api_method(elevation)
#Client.elevation_along_path = make_api_method(elevation_along_path)
#Client.geocode = make_api_method(geocode)
Client.reverse_geocode = make_api_method(reverse_geocode)
#Client.geolocate = make_api_method(geolocate)
#Client.timezone = make_api_method(timezone)
Client.snap_to_roads = make_api_method(snap_to_roads)
#Client.nearest_roads = make_api_method(nearest_roads)
#Client.speed_limits = make_api_method(speed_limits)
#Client.snapped_speed_limits = make_api_method(snapped_speed_limits)
Client.places = make_api_method(places)


def sign_hmac(secret, payload):
    """Returns a base64-encoded HMAC-SHA1 signature of a given string.

    :param secret: The key used for the signature, base64 encoded.
    :type secret: string

    :param payload: The payload to sign.
    :type payload: string

    :rtype: string
    """
    payload = payload.encode('ascii', 'strict')
    secret = secret.encode('ascii', 'strict')
    sig = hmac.new(base64.urlsafe_b64decode(secret), payload, hashlib.sha1)
    out = base64.urlsafe_b64encode(sig.digest())
    return out.decode('utf-8')


def urlencode_params(params):
    """URL encodes the parameters.

    :param params: The parameters
    :type params: list of key/value tuples.

    :rtype: string
    """
    # urlencode does not handle unicode strings in Python 2.
    # Firstly, normalize the values so they get encoded correctly.
    params = [(key, normalize_for_urlencode(val)) for key, val in params]
    # Secondly, unquote unreserved chars which are incorrectly quoted
    # by urllib.urlencode, causing invalid auth signatures. See GH #72
    # for more info.
    return requests.utils.unquote_unreserved(urlencode(params))


try:
    unicode
    # NOTE(cbro): `unicode` was removed in Python 3. In Python 3, NameError is
    # raised here, and caught below.

    def normalize_for_urlencode(value):
        """(Python 2) Converts the value to a `str` (raw bytes)."""
        if isinstance(value, unicode):
            return value.encode('utf8')

        if isinstance(value, str):
            return value

        return normalize_for_urlencode(str(value))

except NameError:
    def normalize_for_urlencode(value):
        """(Python 3) No-op."""
        # urlencode in Python 3 handles all the types we are passing it.
        return value



#
# Copyright 2014 Google Inc. All rights reserved.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Performs requests to the Google Maps Distance Matrix API."""

from neshan import convert


def distance_matrix(client, origins, destinations):
    """ Gets travel distance and time for a matrix of origins and destinations.

    :param origins: One or more locations and/or latitude/longitude values,
        from which to calculate distance and time. If you pass an address as
        a string, the service will geocode the string and convert it to a
        latitude/longitude coordinate to calculate directions.
    :type origins: a single location, or a list of locations, where a
        location is a string, dict, list, or tuple

    :param destinations: One or more addresses and/or lat/lng values, to
        which to calculate distance and time. If you pass an address as a
        string, the service will geocode the string and convert it to a
        latitude/longitude coordinate to calculate directions.
    :type destinations: a single location, or a list of locations, where a
        location is a string, dict, list, or tuple

    :rtype: matrix of distances. Results are returned in rows, each row
        containing one origin paired with each destination.
    """

    params = {
        "origins": convert.location_list(origins),
        "destinations": convert.location_list(destinations)
    }

    return client._request("/v1/distance-matrix", params)
